__author__ = 'Abel Carreras'
__version__ = '0.1.1'

from posym.tools import list_round
from posym.pointgroup import PointGroup
from scipy.spatial.transform import Rotation as R
from scipy.optimize import minimize
import numpy as np
import pandas as pd

cache_orientation = {}
def get_hash(coordinates, symbols, group):
    return hash((np.array2string(coordinates, precision=2),
                 tuple(symbols),
                 group))


class SymmetryBase():
    """
    This class is supposed to be used as a base for more complex symmetry objects

    """
    def __init__(self, group, rep, normalize=False):

        self._pg = PointGroup(group)
        self._group = group.lower()

        if isinstance(rep, str):
            if rep not in self._pg.ir_labels:
                raise Exception('Representation do not match with group\nAvailable: {}'.format(self._pg.ir_labels))
            self._op_representation = self._pg.ir_table[rep]
            #if normalize:
            #    self._op_representation /= self._pg.ir_table[rep]['E']

        elif isinstance(rep, pd.Series):
            if np.all(self._pg.ir_table.sort_index().index == rep.sort_index().index):
                self._op_representation = rep.reindex(self._pg.ir_table.index)
            else:
                raise Exception('Representation not in group')

        if normalize:
            op_rep = np.dot(self._pg.trans_matrix_norm, np.dot(self._pg.trans_matrix_inv, self._op_representation.values))
            self._op_representation = pd.Series(op_rep, index=self._pg.op_labels)

    def get_op_representation(self):
        return self._op_representation

    def get_ir_representation(self):
        ir_rep = np.dot(self._pg.trans_matrix_inv, self._op_representation.values)
        return pd.Series(ir_rep, index=self._pg.ir_labels)

    def get_point_group(self):
        return self._pg

    def __str__(self):

        ir_rep = self.get_ir_representation().values
        ir_rep = list_round(ir_rep)

        ir_labels = self.get_ir_representation().index

        str = ''
        for i, r in enumerate(ir_rep):
            #ir_norm = self._pg.ir_table[ir_labels[i]]['E']
            #r *= ir_norm
            #print('>> ', np.add.reduce(np.square(ir_rep[:i])), 'r:', r)
            if np.add.reduce(np.square(ir_rep[:i])) > 0 and r > 0:
                    str += ' + '
            elif r < 0:
                    str += ' - '
            if np.abs(r - 1) < 2e-2:
                str += ir_labels[i]
            elif np.abs(r) > 0:
                str += '{} {}'.format(abs(r), ir_labels[i])

        return str

    def __repr__(self):
        ir_rep = self.get_ir_representation().values
        ir_rep = list_round(ir_rep)

        ir_labels = self.get_ir_representation().index

        str = ''
        for i, r in enumerate(ir_rep):
            #print('>> ', np.add.reduce(ir_rep[:i]**2), 'r:', r)
            if np.add.reduce(np.square(ir_rep[:i])) > 0 and r > 0 and len(ir_rep[:i]) > 0:
                str += '+'
            if r == 1:
                str += ir_labels[i]
            elif r > 0:
                str += '{}{}'.format(r, ir_labels[i])
        return str

    def __add__(self, other):

        if self._group == other._group:
            return SymmetryBase(self._group,
                                self._op_representation + other._op_representation)

        raise Exception('Incompatible point groups')

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):

        if isinstance(other, (float, int)):
            return SymmetryBase(self._group,
                                self._op_representation * other)

        elif isinstance(other, SymmetryBase):
            mul_rep = self._op_representation * other._op_representation

            return SymmetryBase(self._group, mul_rep)
        else:
            raise Exception('Symmetry operation not possible')


class SymmetryModes(SymmetryBase):
    def __init__(self, group, coordinates, modes, symbols, orientation_angles=None):

        pg = PointGroup(group)

        # set coordinates at geometrical center
        self._coordinates = np.array([c - np.average(coordinates, axis=0) for c in coordinates])

        self._modes = modes
        self._symbols = symbols

        self._coor_measures = []
        self._mode_measures = []

        if orientation_angles is None:
            self._angles = self.get_orientation(pg)
        else:
            self._angles = orientation_angles

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        operations_dic = pg.get_all_operations()

        for operation in pg.operations:
            mode_measures = []
            for op in operations_dic[operation.label]:
                mode_m = op.get_measure(self._coordinates, self._modes, self._symbols, orientation=rotmol)
                mode_measures.append(mode_m)

            mode_measures = np.average(mode_measures, axis=0)
            self._mode_measures.append(mode_measures)

        # De-normalization
        # self._mode_measures = np.dot(pg.trans_matrix, np.dot(pg.trans_matrix_inv_norm, self._mode_measures))

        total_state = pd.Series(np.add.reduce(self._mode_measures, axis=1), index=pg.op_labels)

        super().__init__(group, total_state)

    def get_state_mode(self, n):

        return SymmetryBase(group=self._group, rep=pd.Series(np.array(self._mode_measures).T[n],
                                                             index=self._pg.op_labels))

    def get_orientation(self, pg):

        hash_num = get_hash(self._coordinates, self._symbols, pg.label)
        if hash_num in cache_orientation:
            return cache_orientation[hash_num]

        def optimization_function(angles):

            rotmol = R.from_euler('zyx', angles, degrees=True)

            coor_measures = []
            for operation in pg.operations:
                coor_m = operation.get_measure_pos(self._coordinates, self._symbols, orientation=rotmol)
                coor_measures.append(coor_m)

            # definition group measure
            return np.linalg.norm(coor_measures)

        # preliminar scan
        list_m = []
        list_a = []
        for i in np.arange(0, 180, 36):
            for j in np.arange(0, 180, 36):
                for k in np.arange(0, 180, 36):
                    list_m.append(optimization_function([i, j, k]))
                    list_a.append([i, j, k])

        initial = np.array(list_a[np.nanargmin(list_m)])
        res = minimize(optimization_function, initial, method='CG',
                       # bounds=((0, 360), (0, 360), (0, 360)),
                       # tol=1e-20
                       )
        cache_orientation[hash_num] = res.x
        return cache_orientation[hash_num]

    @property
    def get_measure_pos(self):
        return np.product(self._coor_measures)

    @property
    def opt_coordinates(self):
        rotmol = R.from_euler('zyx', self._angles, degrees=True)
        return rotmol.apply(self._coordinates)

    @property
    def orientation_angles(self):
        return self._angles


class SymmetryFunction(SymmetryBase):
    def __init__(self, group, function, orientation_angles=None, center=None):

        symbols, coordinates = function.get_environment_centers()

        if center is None:
            # center = function.global_center()
            center = np.average(coordinates, axis=0)
            function = function.copy()
            function.apply_translation(-np.array(center))
            coordinates = np.array([c - center for c in coordinates])

        pg = PointGroup(group)

        self._coordinates = np.array(coordinates)

        self._function = function
        self._symbols = symbols

        self._coor_measures = []
        self._operator_measures = []

        if orientation_angles is None:
            self._angles = self.get_orientation(pg)
        else:
            self._angles = orientation_angles

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        self._self_similarity = (self._function * self._function).integrate

        for operation in pg.operations:
            operations_dic = pg.get_all_operations()
            operator_measures = []
            for op in operations_dic[operation.label]:
                operator_m = op.get_measure_func(self._function, self._self_similarity, orientation=rotmol)
                operator_measures.append(operator_m)

            operator_measures = np.average(operator_measures, axis=0)
            self._operator_measures.append(operator_measures)

        # De-normalization
        #self._operator_measures = np.dot(pg.trans_matrix, np.dot(pg.trans_matrix_inv_norm, self._operator_measures))

        total_state = pd.Series(self._operator_measures, index=pg.op_labels)

        super().__init__(group, total_state)

    def get_orientation(self, pg):

        hash_num = get_hash(self._coordinates, self._symbols, pg.label)
        if hash_num in cache_orientation:
            return cache_orientation[hash_num]

        def optimization_function(angles):

            rotmol = R.from_euler('zyx', angles, degrees=True)

            coor_measures = []
            for operation in pg.operations:
                coor_m = operation.get_measure_pos(self._coordinates, self._symbols, orientation=rotmol)
                coor_measures.append(coor_m)

            # definition group measure
            return np.linalg.norm(coor_measures)

        # preliminar scan
        list_m = []
        list_a = []
        for i in np.arange(0, 180, 36):
            for j in np.arange(0, 180, 36):
                for k in np.arange(0, 180, 36):
                    list_m.append(optimization_function([i, j, k]))
                    list_a.append([i, j, k])

        initial = np.array(list_a[np.nanargmin(list_m)])
        res = minimize(optimization_function, initial, method='CG',
                       # bounds=((0, 360), (0, 360), (0, 360)),
                       # tol=1e-20
                       )
        cache_orientation[hash_num] = res.x
        return cache_orientation[hash_num]

    @property
    def self_similarity(self):
        return self._self_similarity

    @property
    def get_measure_pos(self):
        return np.product(self._coor_measures)

    @property
    def opt_coordinates(self):
        rotmol = R.from_euler('zyx', self._angles, degrees=True)
        return rotmol.apply(self._coordinates)

    @property
    def orientation_angles(self):
        return self._angles


if __name__ == '__main__':

    from pyqchem import get_output_from_qchem, Structure
    from pyqchem.tools import get_geometry_from_pubchem

    coordinates = [[ 0.000000000+00,  0.000000000+00,  2.40297090e-01],
                   [-1.43261539e+00, -1.75444785e-16, -9.61188362e-01],
                   [ 1.43261539e+00,  1.75444785e-16, -9.61188362e-01]]

    symbols = ['O', 'H', 'H']

    coordinates = [[0, 0, 0],
                   [ np.sqrt(8/9), 0, -1/3],
                   [-np.sqrt(2/9), np.sqrt(2/3), -1/3],
                   [-np.sqrt(2/9), -np.sqrt(2/3), -1/3],
                   [0, 0, 1]]

    #coordinates = [[ 0.0000000000,    0.0000000000,    0.0000000000],
    #               [ 0.5541000000,    0.7996000000,    0.4965000000],
    #               [ 0.6833000000,   -0.8134000000,   -0.2536000000],
    #               [-0.7782000000,   -0.3735000000,    0.6692000000],
    #               [-0.4593000000,    0.3874000000,   -0.9121000000]]

    coordinates = [[-3.11301739e-06,  1.12541091e-05, -7.97835696e-06],
                   [-6.58614327e-02, -7.77865103e-01, -7.63860353e-01],
                   [-3.64119514e-02,  9.82136003e-01, -4.76386979e-01],
                   [-8.37587195e-01, -9.93456888e-02,  6.93899368e-01],
                   [ 9.39879258e-01, -1.04992740e-01,  5.46395830e-01]]

    symbols = ['C', 'H', 'H', 'H', 'H']


    if True:
        #mol = get_geometry_from_pubchem('methane')
        from pyqchem.file_io import read_structure_from_xyz

        # mol = get_geometry_from_pubchem('Buckminsterfullerene')
        mol = read_structure_from_xyz('../c60.xyz')

        coordinates = mol.get_coordinates()
        symbols = mol.get_symbols()

    from posym.operations.rotation import Rotation, rotation
    from posym.operations.reflection import Reflection, reflection
    from posym.operations.irotation import ImproperRotation
    from posym.operations import get_permutation_simple, get_cross_distance_table

    import matplotlib.pyplot as plt

    if True:
        def optimization_function(angles):
            rotmol = R.from_euler('zyx', angles, degrees=True)
            operation = Rotation(label='C2', axis=[0, 0, 1], order=2)
            # print(operation._axis, operation._order)
            coor_m = operation.get_measure_pos(np.array(coordinates), symbols, orientation=rotmol)
            operation = Reflection(label='s', axis=[0, 1, 0])
            # print(operation._axis, operation._order)
            coor_m2 = operation.get_measure_pos(np.array(coordinates), symbols, orientation=rotmol)
            operation = Rotation(label='C2', axis=[np.sqrt(2/9), 0, 1/3], order=2)
            coor_m3 = operation.get_measure_pos(np.array(coordinates), symbols, orientation=rotmol)
            operation = ImproperRotation(label='S4', axis=[np.sqrt(2/9), 0, 1/3], order=4)
            coor_m4 = operation.get_measure_pos(np.array(coordinates), symbols, orientation=rotmol)

            print(coor_m)
            #print(coor_m + coor_m2 + coor_m3 + coor_m4)
            return coor_m # + coor_m2 + coor_m3 + coor_m4
            #return np.product([coor_m, coor_m2, coor_m3, coor_m4])


        #exit()

        list_m = []
        list_a = []
        for i in np.arange(0, 180, 36):
            for j in np.arange(0, 180, 36):
                for k in np.arange(0, 180, 36):
                    list_m.append(optimization_function([i, j, k]))
                    list_a.append([i, j, k])

        print('------')
        initial = np.array(list_a[np.nanargmin(list_m)])
        #initial = [143.63674808,  -4.31338625,  70.70307201]
        # initial = [43.22189286,  43.86797247, 103.83908054]
        res = minimize(optimization_function, initial, method='CG',
                       #tol=1e20
                       )

        print('res', res.x)
        print('test', optimization_function(res.x))

        # res.x = [43.22189286,  43.86797247, 103.83908054]

        rotmol = R.from_euler('zyx', res.x, degrees=True)

    else:
        angles = [43.22189286,  43.86797247, 103.83908054]
        rotmol = R.from_euler('zyx', angles, degrees=True)


        #print(optimization_function(res.x))
        #rotmol = R.from_euler('zyx', res.x, degrees=True)


    #operation = Rotation(label='C3', axis=[0, 0, 1], order=3)
    #rotmol = R.from_euler('zyx', [0, 0, 0], degrees=True)
    #mode_m, coor_m = operation.get_measure(np.array(coordinates), modes, symbols, orientation=rotmol)
    #print('measure', coor_m)
    #rotmol = R.from_euler('zyx', [360./3, 0, 0], degrees=True)
    #print('---')

    print(Structure(coordinates, symbols))

    # rotations
    rotated_axis = rotmol.apply([0, 0, 1])
    operation = rotation(2*np.pi/2, rotated_axis)
    print('axis', rotated_axis)

    # reflection
    #rotated_axis_r = rotmol.apply([0, 1,  0])
    #operation = reflection(rotated_axis_r)
    #print('axis_r', rotated_axis_r)

    # C2 rotation
    # rotated_axis_i = rotmol.apply([(np.sqrt(8 / 9))/2, 0, (1-1 / 3)/2])

    # S4 rotation
    #rotated_axis_i = rotmol.apply([np.sqrt(2 / 9), 0, 1 / 3])
    #operation1 = rotation(2*np.pi/4, rotated_axis_i)
    #operation2 = reflection(rotated_axis_i)
    #operation = np.dot(operation2, operation1)
    #print('axis_i', rotated_axis_i)

    #coordinates = rotmol.apply(coordinates)
    #print(Structure(coordinates, symbols))

    print('det:', np.linalg.det(operation))
    permu_coor = np.dot(operation, np.array(coordinates).T).T

    print(Structure(permu_coor, symbols))


    distance_table = get_cross_distance_table(coordinates, permu_coor)
    perm = get_permutation_simple(distance_table, symbols)
    #print(np.round(distance_table[perm], 3))
    print(perm)

    permu_coor = permu_coor[perm]

    print(Structure(permu_coor, symbols))

    print('****')
    print(permu_coor)
    a = np.average(np.linalg.norm(np.subtract(coordinates, permu_coor), axis=0))
    print(a)



# algebra opearation functions for states
import numpy as np


def dot(state1, state2, normalize=False):

    if state1.get_point_group() != state2.get_point_group():
        raise ValueError("States must have same point group")

    pg = state1.get_point_group()
    norm = pg.ir_table.T['E'].values

    v1 = state1.get_ir_representation()
    v2 = state2.get_ir_representation()

    dot = np.sum([np.sqrt((a*b).clip(min=0)) for a, b in zip(np.multiply(v1.values, norm),
                                                             np.multiply(v2.values, norm))])

    if normalize:

        n1 = np.sum(np.multiply(v1.values, norm))
        n2 = np.sum(np.multiply(v2.values, norm))

        dot = dot / np.sqrt(n1 * n2)

    dot = np.square(dot)

    return dot


def norm(state1):

    v1 = state1.get_ir_representation()
    pg = state1.get_point_group()
    norm = pg.ir_table.T['E'].values

    dot = np.sum([a*n for a, n in zip(v1.values, norm)])

    return dot


# This is under contruction and not yet in use

import numpy as np
from math import gcd
import pandas as pd
from posym.operations.identity import Identity
from posym.operations.rotation import Rotation
from posym.operations.reflection import Reflection
from posym.operations.inversion import Inversion
from posym.operations.irotation import ImproperRotation
from posym.ir_tables import CharTable
from posym.ir_tables import real_radical
from posym.ir_tables import ir_table_list


def get_cn(n):

    r = np.clip(np.mod(n-1, 2), 0, 1)
    l = (n-r-1)//2
    ndim = 1 + l + r

    multiplicity = [1] + [2] * l + [1] * r

    operations = [Identity(label='E')]
    for i in range(n//2):
        j = gcd(i+1, n)
        up = (i+1)//j
        down = n//j
        label = 'C{}'.format(down) if up == 1 else 'C^{}_{}'.format(up, down)

        operations.append(Rotation(label=label, axis=[0, 0, 1], order=down, exp=up))

    ir_data = {'A': pd.Series([1] * ndim)}
    if r == 1:
        ir_data.update({'B': pd.Series([(-1)**k for k in range(ndim)])})

    for i in range(l):
        label = 'E{}'.format(i+1) if l > 1 else 'E'
        ir_data.update({label: pd.Series([2] + [real_radical((i+1)*(k+1), n) for k in range(ndim-1)])})

    if n == 1:
        rotations = ['A', 'A', 'A']
        translations = ['A', 'A', 'A']
    elif n == 2:
        rotations = ['B', 'B', 'A']
        translations = ['B', 'B', 'A']
    elif n == 3 or n == 4:
        rotations = ['E', 'E', 'A']
        translations = ['E', 'E', 'A']
    else:
        rotations = ['E1', 'E1', 'A']
        translations = ['E1', 'E1', 'A']

    name = 'C{}'.format(n)
    #return name, operations, ir_data, rotations, translations, multiplicity
    return CharTable('C{}'.format(n),
                     operations,
                     ir_data,
                     rotations=rotations,  # x, y, z
                     translations=translations,  # Rx, Ry, Rz
                     multiplicities=multiplicity)


def get_cnv(n):

    if n<2:
        raise Exception('Group not valid')

    ir_data = get_cn(n)
    operations_new = ir_data.operations

    ir_data_new = {}
    ir_data_new['A1'] = pd.Series(list(ir_data['A']) + [ 1])
    ir_data_new['A2'] = pd.Series(list(ir_data['A']) + [-1])
    operations_new += [Reflection(label='sv_xz', axis=[0, 1, 0])]

    if 'B' in ir_data:
        ir_data_new['B1'] = pd.Series(list(ir_data['B']) + [ 1, -1])
        ir_data_new['B2'] = pd.Series(list(ir_data['B']) + [-1,  1])

        ir_data_new['A1'] = pd.Series(list(ir_data_new['A1']) + [ 1])
        ir_data_new['A2'] = pd.Series(list(ir_data_new['A2']) + [-1])

        operations_new += [Reflection(label='sd_yz', axis=[1, 0, 0])]

    for data in ir_data.keys():
        if data.startswith('E'):
            ir_data_new[data] = pd.Series(list(ir_data[data]) + [0])

    if 'B' in ir_data:
        for data in ir_data.keys():
            if data.startswith('E'):
                ir_data_new[data] = pd.Series(list(ir_data_new[data]) + [0])

    multiplicites = ir_data.multiplicities
    if np.mod(n, 2) == 0:
        multiplicites += [n//2, n//2]
    else:
        multiplicites += [n]

    if n == 2:
        rotations = ['B2', 'B2', 'A1']
        translations = ['B1', 'B2', 'A1']
    elif n == 3 or n == 4:
        rotations = ['E', 'E', 'A2']
        translations = ['E', 'E', 'A1']
    else:
        rotations = ['E1', 'E1', 'A2']
        translations = ['E1', 'E1', 'A1']

    return CharTable('C{}v'.format(n),
                     operations_new,
                     ir_data_new,
                     rotations=rotations,  # x, y, z
                     translations=translations,  # Rx, Ry, Rz
                     multiplicities=multiplicites)


def get_cnh(n):

    ir_data = get_cn(n)
    operations_new = ir_data.operations
    ir_data_new = {}
    ir_data_new_u = {}

    if np.mod(n, 2) == 0:
        # for even n only

        ir_data_new['Ag'] = pd.Series(list(ir_data['A']) + [ 1, 1])
        ir_data_new_u['Au'] = pd.Series(list(ir_data['A']) + [-1, -1])

        l = (-1)**(n//2)
        ir_data_new['Bg'] = pd.Series(list(ir_data['B']) + [ 1, 1*l])
        ir_data_new_u['Bu'] = pd.Series(list(ir_data['B']) + [-1, -1*l])

        operations_new += [Inversion(label='i'), Reflection(label='sh', axis=[0, 0, 1])]

        for data in ir_data.keys():
            if data.startswith('E'):
                if len(data[1:]) == 0:
                    k = int((np.mod(1, 2) -0.5)*4)
                else:
                    k = int((np.mod(int(data[1:]), 2) -0.5)*4)
                # print(data+'u', list(ir_data[data]))
                ir_data_new[data+'g'] = pd.Series(list(ir_data[data]) + [2, -k])
                ir_data_new_u[data+'u'] = pd.Series(list(ir_data[data]) + [-2, k])

        ir_data_new.update(ir_data_new_u)

        for i in range((n-1)//2):
            j = gcd(i+1, n)
            up = (i+1)//j
            down = n//j
            if up == 1:
                operations_new += [ImproperRotation(label='S_{}'.format(down), axis=[0, 0, 1], order=down, exp=up)]
            else:
                operations_new += [ImproperRotation(label='S^{}_{}'.format(up, down), axis=[0, 0, 1], order=down, exp=up)]

            for data in ir_data_new.keys():
                element = list(ir_data_new[data])[1+i] * np.sign(list(ir_data_new[data])[(n-1)//2+3])
                ir_data_new[data] = pd.Series(list(ir_data_new[data]) + [element])

        l = 2 + n // 2
        # swap columns for convention
        value = operations_new.pop(l)
        operations_new.append(value)
        for data in ir_data_new.keys():
            row = list(ir_data_new[data])
            value = row.pop(l)
            row.append(value)
            ir_data_new[data] = pd.Series(row)


        multiplicites = [1] + [2]*((n-1)//2) + [1, 1] + [2]*((n-1)//2) + [1]

        if n == 2:
            rotations = ['Bu', 'Bu', 'Au']
            translations = ['Bg', 'Bg', 'Ag']
        elif  n == 4:
            rotations = ['Eu', 'Eu', 'Au']
            translations = ['Eg', 'Eg', 'Ag']
        else:
            rotations = ['E1u', 'E1u', 'Au']
            translations = ['E1g', 'E1g', 'Ag']
    else:

        ir_data_new["A'"] = pd.Series(list(ir_data['A']) + [1])
        ir_data_new["A''"] = pd.Series(list(ir_data['A']) + [-1])

        multiplicites = [1] + [2]*((n-1)//2) + [1] + [2]*((n-1)//2)

        operations_new += [Reflection(label='sh', axis=[0, 0, 1])]

        for data in ir_data.keys():
            if data.startswith('E'):
                ir_data_new[data+"'"] = pd.Series(list(ir_data[data]) + [2])
                ir_data_new[data+"''"] = pd.Series(list(ir_data[data]) + [-2])

        for i in range((n-1)//2):
            j = gcd(2*i+1, n)
            up = (2*i+1)//j
            #print('----------------', n, i*2+1, ':', (2*i+1)//j, n//j)
            down = n//j
            if up == 1:
                operations_new += [ImproperRotation(label='S_{}'.format(down), axis=[0, 0, 1], order=down, exp=up)]
            else:
                operations_new += [ImproperRotation(label='S^{}_{}'.format(up, down), axis=[0, 0, 1], order=down, exp=up)]

            for l, data in enumerate(ir_data_new.keys()):
                # real_radical((i + 1) * (k + 1), n)
                sign = np.sign(list(ir_data_new[data])[(n - 1) // 2 + 1])
                element = list(ir_data_new[data])[1+i] * sign
                if data.startswith('E'):
                    try:
                        k = int(data[1])-1
                    except ValueError:
                        k = 0

                    # print([real_radical((k + 1) * (m + 1), n) for m in range(0, n,2)])
                    # element = [real_radical((k + 1) * (m + 1), n) for m in range(0, n, 2)][i] * sign
                    element = real_radical((k + 1) * (2*i + 1), n) * sign

                ir_data_new[data] = pd.Series(list(ir_data_new[data]) + [element])

        if n == 1:
            rotations = ["A'", "A'", "A''"]
            translations = ["A''", "A''", "A'"]
        elif n == 3:
            rotations = ["E'", "E'", "A''"]
            translations = ["E''", "E''", "A'"]
        else:
            rotations = ["E1'", "E1'", "A''"]
            translations = ["E1''", "E1''", "A'"]

    return CharTable('C{}h'.format(n),
                     operations_new,
                     ir_data_new,
                     rotations=rotations,  # x, y, z
                     translations=translations,  # Rx, Ry, Rz
                     multiplicities=multiplicites)

def get_sn(n):

    if np.mod(n, 2) != 0:
        raise Exception('Order of Sn group must have even')

    if np.mod(n, 4) == 0:

        r = np.clip(np.mod(n-1, 2), 0, 1)
        l = (n-r-1)//2
        ndim = 1 + l + r

        multiplicity = [1] + [2] * l + [1] * r

        operations = [Identity(label='E')]
        for i in range(n//2):
            j = gcd(i+1, n)
            up = (i+1)//j
            down = n//j
            if np.mod(i, 2) == 0:
                label = 'S{}'.format(down) if up == 1 else 'S^{}_{}'.format(up, down)
                operations.append(ImproperRotation(label=label, axis=[0, 0, 1], order=down, exp=up))
            else:
                label = 'C{}'.format(down) if up == 1 else 'C^{}_{}'.format(up, down)
                operations.append(Rotation(label=label, axis=[0, 0, 1], order=down, exp=up))

        ir_data = {'A': pd.Series([1] * ndim)}
        if r == 1:
            ir_data.update({'B': pd.Series([(-1)**k for k in range(ndim)])})

        for i in range(l):
            label = 'E{}'.format(i+1) if l > 1 else 'E'
            ir_data.update({label: pd.Series([2] + [real_radical((i+1)*(k+1), n) for k in range(ndim-1)])})

        if n == 4:
            rotations = ['E', 'E', 'A']
            translations = ['E', 'E', 'B']
        else:
            rotations = ['E1', 'E1', 'A']
            translations = ['E{}'.format(l//2), 'E{}'.format(l//2), 'B']

    else:

        r = np.clip(np.mod(n-1, 2), 0, 1)
        l = (n-r-1)//2
        ndim = 1 + l + r

        multiplicity = [1] + [2] * l + [1] * r

        operations = [Identity(label='E')]
        for i in range(n//2):
            j = gcd(i+1, n)
            up = (i+1)//j
            down = n//j
            if np.mod(i, 2) == 0:
                if up == 1 and down == 2:
                    label = 'i'.format(down)
                    operations.append(Inversion(label=label))
                else:
                    label = 'S{}'.format(down) if up == 1 else 'S^{}_{}'.format(up, down)
                    operations.append(ImproperRotation(label=label, axis=[0, 0, 1], order=down, exp=up))
            else:
                label = 'C{}'.format(down) if up == 1 else 'C^{}_{}'.format(up, down)
                operations.append(Rotation(label=label, axis=[0, 0, 1], order=down, exp=up))

        ir_data = {'Ag': pd.Series([1] * ndim)}
        traces = [1]
        for i in range(n//2):
            if np.mod(i, 2) == 0:
                traces.append(-1)
            else:
                traces.append(1)

        for i in range(l//2):
            subtrace = np.roll([k+1 for k in range(n//4)], -i)
            label = 'E{}g'.format(i+1) if l > 2 else 'Eg'
            ir_data.update({label: pd.Series([2] + [real_radical(k, n)*(-1)**(k+2*i) for k in subtrace] + [real_radical(k, n)*(-1)**(k+2*i) for k in subtrace[::-1]] + [2])})

        ir_data.update({'Au': pd.Series(traces)})

        for i in range(l//2):
            subtrace = np.roll([k + 1 for k in range(n // 4)], -i)
            label = 'E{}u'.format(i+1) if l > 2 else 'Eu'
            # print(i, '-', [[2] , [(-1)**(m+1) for m, k in enumerate(subtrace)] , [(-1)**(m+l//2+1) for m, k in enumerate(subtrace[::-1])] , [-2]])
            ir_data.update({label: pd.Series([2] +
                                             [real_radical(k, n)*(-1)**(k+2*i) * (-1)**(m+1) for m, k in enumerate(subtrace)] +
                                             [real_radical(k, n)*(-1)**(k+2*i) *(-1)**(m+l//2+1) for m, k in enumerate(subtrace[::-1])] +
                                             [-2])})

        # reorder columns to follow convention
        new_indexing = [0] + list(range(2, l+1, 2)) + [ndim-1] + list(range(1, l+1, 2))
        operations = np.array(operations)[new_indexing].tolist()
        multiplicity = np.array(multiplicity)[new_indexing].tolist()
        for key in ir_data.keys():
            ir_data[key] = ir_data[key][new_indexing]

        if n == 2:
            rotations = ['Ag', 'Ag', 'Ag']
            translations = ['Au', 'Au', 'Au']
        elif n == 6:
            rotations = ['Eg', 'Eg', 'Ag']
            translations = ['Eu', 'Eu', 'Au']
        else:
            rotations = ['E1g', 'E1g', 'Ag']
            translations = ['E1u', 'E1u', 'Au']

    return CharTable('S{}'.format(n),
                     operations,
                     ir_data,
                     rotations=rotations,  # x, y, z
                     translations=translations,  # Rx, Ry, Rz
                     multiplicities=multiplicity)


def get_dn(n):

    if n<2:
        raise Exception('Group not valid')

    if n == 2:

        return CharTable('D2',
                         [Identity(label='E'), Rotation(label='C2', axis=[0, 0, 1], order=2),
                          Rotation(label="C2'", axis=[0, 1, 0]), Reflection(label="C2''", axis=[1, 0, 0])],
                         {'A1': pd.Series([+1, +1, +1, +1]),
                          'B1': pd.Series([+1, +1, -1, -1]),
                          'B2': pd.Series([+1, -1, -1, +1]),
                          'B3': pd.Series([+1, -1, +1, -1])
                         },
                         rotations=['B3', 'B2', 'B1'],
                         translations=['B3', 'B2', 'B1'],
                         multiplicities=[1, 1, 1, 1])

    ir_data = get_cn(n)
    operations_new = ir_data.operations

    ir_data_new = {}
    ir_data_new['A1'] = pd.Series(list(ir_data['A']) + [ 1])
    ir_data_new['A2'] = pd.Series(list(ir_data['A']) + [-1])
    operations_new += [Rotation(label="C2'", axis=[0, 1, 0])]

    if 'B' in ir_data:
        ir_data_new['B1'] = pd.Series(list(ir_data['B']) + [ 1, -1])
        ir_data_new['B2'] = pd.Series(list(ir_data['B']) + [-1,  1])

        ir_data_new['A1'] = pd.Series(list(ir_data_new['A1']) + [ 1])
        ir_data_new['A2'] = pd.Series(list(ir_data_new['A2']) + [-1])

        operations_new += [Rotation(label="C2''", axis=[1, 0, 0])]

    for data in ir_data.keys():
        if data.startswith('E'):
            ir_data_new[data] = pd.Series(list(ir_data[data]) + [0])

    if 'B' in ir_data:
        for data in ir_data.keys():
            if data.startswith('E'):
                ir_data_new[data] = pd.Series(list(ir_data_new[data]) + [0])

    multiplicites = ir_data.multiplicities
    if np.mod(n, 2) == 0:
        multiplicites += [n//2, n//2]
    else:
        multiplicites += [n]

    if n == 2:
        rotations = ['B2', 'B2', 'A1']
        translations = ['B1', 'B2', 'A1']
    elif n == 3 or n == 4:
        rotations = ['E', 'E', 'A2']
        translations = ['E', 'E', 'A2']
    else:
        rotations = ['E1', 'E1', 'A2']
        translations = ['E1', 'E1', 'A2']

    return CharTable('D{}'.format(n),
                     operations_new,
                     ir_data_new,
                     rotations=rotations,  # x, y, z
                     translations=translations,  # Rx, Ry, Rz
                     multiplicities=multiplicites)


def get_dnh(n):

    ir_data = get_dn(n)
    operations_new = ir_data.operations
    ir_data_new = {}
    ir_data_new_u = {}

    if n == 2:
        return CharTable('D2h',
                         [Identity(label='E'), Rotation(label='C2', axis=[0, 0, 1], order=2),
                          Rotation(label="C2'", axis=[0, 1, 0]), Rotation(label="C2''", axis=[1, 0, 0]),
                          Inversion(label='i'), Reflection(label='sh', axis=[0, 0, 1]),
                          Reflection(label='sv', axis=[0, 1, 0]), Reflection(label='sd', axis=[1, 0, 0])],
                         {'Ag': pd.Series( [+1, +1, +1, +1, +1, +1, +1, +1]),
                          'B1g': pd.Series([+1, +1, -1, -1, +1, +1, -1, -1]),
                          'B2g': pd.Series([+1, -1, -1, +1, +1, -1, +1, -1]),
                          'B3g': pd.Series([+1, -1, +1, -1, +1, -1, -1, +1]),
                          'Au': pd.Series( [+1, +1, +1, +1, -1, -1, -1, -1]),
                          'B1u': pd.Series([+1, +1, -1, -1, -1, -1, +1, +1]),
                          'B2u': pd.Series([+1, -1, -1, +1, -1, +1, -1, +1]),
                          'B3u': pd.Series([+1, -1, +1, -1, -1, +1, +1, -1]),
                          },
                         rotations=['B3g', 'B2g', 'B1g'],
                         translations=['B3u', 'B2u', 'B1u'],
                         multiplicities=[1, 1, 1, 1, 1, 1, 1, 1])

    if np.mod(n, 2) == 0:
        # for even n only

        ir_data_new['A1g'] =   pd.Series(list(ir_data['A1']) + [ 1,  1,  1,  1])
        ir_data_new_u['A1u'] = pd.Series(list(ir_data['A1']) + [-1, -1, -1, -1])

        ir_data_new['A2g'] =   pd.Series(list(ir_data['A2']) + [ 1,  1, -1, -1])
        ir_data_new_u['A2u'] = pd.Series(list(ir_data['A2']) + [-1, -1,  1,  1])

        l = (-1)**(n//2)
        ir_data_new['B1g'] =   pd.Series(list(ir_data['B1']) + [ 1,  1*l,  1*l, -1*l])
        ir_data_new_u['B1u'] = pd.Series(list(ir_data['B1']) + [-1, -1*l, -1*l,  1*l])

        ir_data_new['B2g'] =   pd.Series(list(ir_data['B2']) + [ 1,  1*l,  -1*l,  1*l])
        ir_data_new_u['B2u'] = pd.Series(list(ir_data['B2']) + [-1, -1*l,   1*l, -1*l])

        operations_new += [Inversion(label='i'), Reflection(label='sh', axis=[0, 0, 1]),
                           Reflection(label='s_v', axis=[0, 1, 0]), Reflection(label='sd', axis=[1, 0, 0])]

        for data in ir_data.keys():
            if data.startswith('E'):
                if len(data[1:]) == 0:
                    k = int((np.mod(1, 2) -0.5)*4)
                else:
                    k = int((np.mod(int(data[1:]), 2) -0.5)*4)
                # print(data+'u', list(ir_data[data]))
                ir_data_new[data+'g'] = pd.Series(list(ir_data[data]) + [2, -k, 0, 0])
                ir_data_new_u[data+'u'] = pd.Series(list(ir_data[data]) + [-2, k, 0, 0])

        ir_data_new.update(ir_data_new_u)

        for i in range((n-1)//2):
            j = gcd(i+1, n)
            up = (i+1)//j
            down = n//j
            if up == 1:
                operations_new += [ImproperRotation(label='S{}'.format(down), axis=[0, 0, 1], order=down, exp=up)]
            else:
                operations_new += [ImproperRotation(label='S^{}_{}'.format(up, down), axis=[0, 0, 1], order=down, exp=up)]

            for data in ir_data_new.keys():
                #print(1+i, (n-1)//2+5)
                element = list(ir_data_new[data])[1+i] * np.sign(list(ir_data_new[data])[(n-1)//2+5])
                #print(data, len(ir_data_new[data]), element)
                ir_data_new[data] = pd.Series(list(ir_data_new[data]) + [element])
            #print('----')

        # swap columns for convention
        l = 2 + n // 2+2
        for kk in range(3):
            value = operations_new.pop(l)
            operations_new.append(value)
            for data in ir_data_new.keys():
                row = list(ir_data_new[data])
                value = row.pop(l)
                row.append(value)
                ir_data_new[data] = pd.Series(row)

        #print(((n-1)//2))
        multiplicites = [1] + [2]*((n-1)//2) + [1]  + [n//2, n//2] + [1] + [2]*((n-1)//2) + [1, n//2, n//2]

        print(multiplicites)
        if n == 2:
            rotations = ['B3g', 'B2g', 'B1g']
            translations = ['B3u', 'B2u', 'B1u']
        elif  n == 4:
            rotations = ['Eg', 'Eg', 'A2g']
            translations = ['Eu', 'Eu', 'A2u']
        else:
            rotations = ['E1g', 'E1g', 'A2g']
            translations = ['E1u', 'E1u', 'A2u']
    else:

        ir_data_new["A1'"] = pd.Series(list(ir_data['A1']) + [1, 1])
        ir_data_new["A1''"] = pd.Series(list(ir_data['A1']) + [-1, -1])

        ir_data_new["A2'"] = pd.Series(list(ir_data['A2']) + [1, -1])
        ir_data_new["A2''"] = pd.Series(list(ir_data['A2']) + [-1, 1])

        multiplicites = [1] + [2]*((n-1)//2) + [n, 1] + [2]*((n-1)//2) + [n]

        operations_new += [Reflection(label='sh', axis=[0, 0, 1]),
                           Reflection(label='sv', axis=[0, 1, 0])]

        for data in ir_data.keys():
            if data.startswith('E'):
                ir_data_new[data+"'"] = pd.Series(list(ir_data[data]) + [2, 0])
                ir_data_new[data+"''"] = pd.Series(list(ir_data[data]) + [-2, 0])

        for i in range((n-1)//2):
            j = gcd(2*i+1, n)
            up = (2*i+1)//j
            down = n//j
            if up == 1:
                operations_new += [ImproperRotation(label='S_{}'.format(down), axis=[0, 0, 1], order=down, exp=up)]
            else:
                operations_new += [ImproperRotation(label='S^{}_{}'.format(up, down), axis=[0, 0, 1], order=down, exp=up)]

            for data in ir_data_new.keys():
                sign = np.sign(list(ir_data_new[data])[(n - 1) // 2 + 2])
                element = list(ir_data_new[data])[1+i] * sign
                if data.startswith('E'):
                    try:
                        k = int(data[1])-1
                    except ValueError:
                        k = 0

                    # print('E{}'.format(k+1), [real_radical((k + 1) * (m + 1), n) for m in range(0, 7)])
                    # element = [real_radical((k + 1) * (m + 1), n) for m in range(0, n,2)][i] * sign
                    element = real_radical((k + 1) * (2*i + 1), n) * sign

                ir_data_new[data] = pd.Series(list(ir_data_new[data]) + [element])

        # swap columns for convention
        l = 2 + n // 2+1
        for kk in range(1):
            value = operations_new.pop(l)
            operations_new.append(value)
            for data in ir_data_new.keys():
                row = list(ir_data_new[data])
                value = row.pop(l)
                row.append(value)
                ir_data_new[data] = pd.Series(row)

        if n == 1:
            rotations = ["A1''", "A2''", "A2'"]
            translations = ["A1'", "A2'", "A2''"]
        elif n == 3:
            rotations = ["E''", "E''", "A2'"]
            translations = ["E'", "E'", "A2''"]
        else:
            rotations = ["E1''", "E1''", "A2'"]
            translations = ["E1'", "E1'", "A2''"]

    return CharTable('D{}h'.format(n),
                     operations_new,
                     ir_data_new,
                     rotations=rotations,  # x, y, z
                     translations=translations,  # Rx, Ry, Rz
                     multiplicities=multiplicites)

def get_dnd(n):

    ir_data_new = {}

    if n < 2:
        raise Exception('Group not valid')

    if np.mod(n, 2) == 0:
        # for even n only
        ir_data = get_sn(2*n)
        operations_new = ir_data.operations

        ir_data_new['A1'] = pd.Series(list(ir_data['A']) + [ 1,  1])
        ir_data_new['A2'] = pd.Series(list(ir_data['A']) + [-1, -1])

        ir_data_new['B1'] = pd.Series(list(ir_data['B']) + [ 1, -1])
        ir_data_new['B2'] = pd.Series(list(ir_data['B']) + [-1,  1])

        operations_new += [Rotation(label="C2'", axis=[1, 0, 0], order=2), Reflection(label='sd', axis=[1, 0, 0])]

        for data in ir_data.keys():
            if data.startswith('E'):
                ir_data_new[data] = pd.Series(list(ir_data[data]) + [0, 0])

        multiplicites = ir_data.multiplicities + [n, n]

        if n == 2:
            rotations = ['E', 'E', 'A2']
            translations = ['E', 'E', 'B2']
        else:
            rotations = ['E{}'.format(n-1), 'E{}'.format(n-1), 'A2']
            translations = ['E1', 'E1', 'B2']
    else:
        ir_data = get_dn(n)
        operations_new = ir_data.operations
        ir_data_new_u = {}

        ir_data_new["A1g"] = pd.Series(list(ir_data['A1']) + [1, 1])
        ir_data_new["A2g"] = pd.Series(list(ir_data['A2']) + [1, -1])

        ir_data_new_u["A1u"] = pd.Series(list(ir_data['A1']) + [-1, -1])
        ir_data_new_u["A2u"] = pd.Series(list(ir_data['A2']) + [-1, 1])

        multiplicites = [1] + [2]*((n-1)//2) + [n, 1] + [2]*((n-1)//2) + [n]

        operations_new += [Inversion(label='i'),
                           Reflection(label='sd', axis=[1, 0, 0])]

        for data in ir_data.keys():
            if data.startswith('E'):
                ir_data_new[data+"g"] = pd.Series(list(ir_data[data]) + [2, 0])
                ir_data_new_u[data+"u"] = pd.Series(list(ir_data[data]) + [-2, 0])

        ir_data_new.update(ir_data_new_u)

        for i in range((n-1)//2):
            j = gcd(2*i+1, 2*n)
            up = (2*i+1)//j
            down = 2*n//j
            if up == 1:
                operations_new += [ImproperRotation(label='S{}'.format(down), axis=[0, 0, 1], order=down, exp=up)]
            else:
                operations_new += [ImproperRotation(label='S^{}_{}'.format(up, down), axis=[0, 0, 1], order=down, exp=up)]

            for data in ir_data_new.keys():
                sign = np.sign(list(ir_data_new[data])[(n - 1) // 2 + 2])
                element = list(ir_data_new[data])[1+i] * sign
                if data.startswith('E'):
                    try:
                        k = int(data[1])-1
                    except ValueError:
                        k = 0

                    #element = [real_radical((k + 1) * (m + 1), n) for m in range(0, n//2)][-i-1] * sign
                    element = real_radical((k + 1) * (n // 2 + i + 1), n) * sign

                ir_data_new[data] = pd.Series(list(ir_data_new[data]) + [element])

        # swap columns for convention
        l = 2 + n // 2+1
        for kk in range(1):
            value = operations_new.pop(l)
            operations_new.append(value)
            for data in ir_data_new.keys():
                row = list(ir_data_new[data])
                value = row.pop(l)
                row.append(value)
                ir_data_new[data] = pd.Series(row)

        if n == 1:
            rotations = ["A1g", "A2g", "A2g"]
            translations = ["A1'", "A2'", "A2''"]
        elif n == 3:
            rotations = ["Eg", "Eg", "A2g"]
            translations = ["Eu", "Eu", "A2u"]
        else:
            rotations = ["E1g", "E1g", "A2g"]
            translations = ["E1u", "E1u", "A2u"]

    return CharTable('D{}d'.format(n),
                     operations_new,
                     ir_data_new,
                     rotations=rotations,  # x, y, z
                     translations=translations,  # Rx, Ry, Rz
                     multiplicities=multiplicites)


def get_table_from_label(label):

    label = label.lower().strip()

    # Check into explicitly defined tables
    for table in ir_table_list:
        if label == table.name.lower():
            return table

    # Generate table
    type = label[0]
    if type in ['c', 'd', 's']:
        code = type[0]
        subtype = label[-1]
        try:
            if subtype in ['d', 'h', 'v']:
                order = int(label[1:-1])
                code += subtype
            else:
                order = int(label[1:])
        except ValueError:
            code = order = None

        functions = {'d': get_dn, 'dd': get_dnd, 'dh': get_dnh,
                     'c': get_cn, 'ch': get_cnh, 'cv':get_cnv,
                     's': get_sn}

        try:
            return functions[code](order)
        except KeyError:
            pass

    raise Exception('Point group label {} not recognized'.format(label))


if __name__ == '__main__':

    pg = get_table_from_label('c1')
    print(pg.name)
    print(pg)
    exit()
    #for i in range(1, 9):
    #    print(get_cnh(i))

    print(get_dnd(2))
    print(get_dnd(3))
    print(get_dnd(4))
    print(get_dnd(5))
    print(get_dnd(6))
    print(get_dnd(7))
    print(get_dnd(8))

    exit()

    print(get_dnh(3))
    print(get_dnh(5))
    print(get_dnh(7))
    print(get_dnh(9))

    exit()
    print(get_dn(5))
    print(get_dnh(5))
    print(get_dnh(6))
    print(get_dnh(7))

    exit()



import numpy as np
from posym.generate_tables import get_table_from_label


class PointGroup():
    """
    Basic class to basically handle IR-table.

    """
    def __init__(self, group):
        self._group = group.lower()

        self._trans_matrix = None
        self._trans_matrix_inv = None
        self._trans_matrix_norm = None
        self._trans_matrix_inv_norm = None

        # get IR table
        self._table = get_table_from_label(group)

    def __hash__(self):
        return hash(self._group)

    def __eq__(self, other):
        return self.label == other.label

    def __str__(self):
        return '{}\n{}'.format(self._table.name, self._table.T.to_string())

    def get_all_operations(self):
        return self._table.get_all_operations()

    @property
    def group(self):
        return self._group

    @property
    def operations(self):
        return self._table.operations

    @property
    def ir_table(self):
        return self._table

    @property
    def op_labels(self):
        return self._table.index

    @property
    def ir_labels(self):
        return self._table.T.index

    @property
    def n_sym_elements(self):
        return np.sum(self._table.multiplicities)

    @property
    def n_ir(self):
        return len(self._table)

    @property
    def trans_matrix(self):
        """
        transforms IR to Op

        Op  = Mat * IR

        :return: the transformation matrix
        """
        if self._trans_matrix is None:
            self._trans_matrix = np.array([v for v in self.ir_table.T.values]).T
            self._trans_matrix = self._trans_matrix

        return self._trans_matrix

    @property
    def trans_matrix_norm(self):
        """
        transforms IR to Op

        Op  = Mat * IR

        :return: the transformation matrix
        """
        if self._trans_matrix_norm is None:
            self._trans_matrix_norm = np.array([v for v in self.ir_table.T.values]).T
            self._trans_matrix_norm = self._trans_matrix_norm/self._trans_matrix_norm[0, :]  # normalization

        return self._trans_matrix_norm

    @property
    def trans_matrix_inv(self):
        """
        transforms Op to IR

        IR = Mat * Op

        :return: the transformation matrix
        """

        if self._trans_matrix_inv is None:
            self._trans_matrix_inv = np.linalg.inv(self.trans_matrix)

        return self._trans_matrix_inv

    @property
    def trans_matrix_inv_norm(self):
        """
        transforms Op to IR

        IR = Mat * Op

        :return: the transformation matrix
        """

        if self._trans_matrix_inv_norm is None:
            self._trans_matrix_inv_norm = np.linalg.inv(self.trans_matrix_norm)

        return self._trans_matrix_inv_norm

    @property
    def order(self):
        return np.sum(self._table.multiplicities)

    @property
    def ir_degeneracies(self):
        return self._table.ir_degeneracies

    @property
    def label(self):
        return self._group
